use nostr_rs_client::{Clients, Error as ClientError};
use nostr_rs_types::{
    client::{Close, Subscribe},
    types::{Addr, Content, Event, Filter, Kind, SubscriptionId, Tag},
    Response,
};
use parking_lot::RwLock;
use sqlx::{query, FromRow, Pool, Sqlite, SqlitePool};
use std::{collections::HashMap, sync::Arc};
use tokio::time::{sleep, Duration};

#[derive(Clone, FromRow, Debug)]
struct ToFetch {
    pub id: String,
    pub typ: i64,
}

#[derive(Clone, FromRow, Debug)]
struct Relayer {
    pub id: i64,
    pub url: String,
}

#[derive(Clone, FromRow, Debug)]
struct RawValue {
    pub id: String,
    pub event: String,
}

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("Sql: {0}")]
    Sql(#[from] sqlx::Error),

    #[error("Nostr: {0}")]
    Addr(#[from] nostr_rs_types::types::addr::Error),

    #[error("client: {0}")]
    Client(#[from] ClientError),
}

async fn discover_relayers(conn: Pool<Sqlite>) -> Result<(), Error> {
    let mut tx = conn.begin().await?;
    let values = sqlx::query_as::<_, RawValue>(
        r#"SELECT id, event FROM events WHERE processed = 1 LIMIT 10000"#,
    )
    .fetch_all(&mut tx)
    .await?;

    let mut relayers: HashMap<String, u64> = HashMap::new();

    for value in values {
        let event: Result<Event, _> = serde_json::from_str(&value.event);
        match event {
            Ok(event) => {
                match event.content() {
                    Content::Contacts(x) => x
                        .iter()
                        .map(|(relayer, _)| {
                            let relayer = relayer.to_ascii_lowercase();
                            if !relayer.is_empty() {
                                if relayers.get_mut(&relayer).map(|x| *x += 1).is_none() {
                                    relayers.insert(relayer, 1);
                                }
                            }
                        })
                        .for_each(drop),
                    _ => {}
                }
                for tag in event.inner.tags.iter() {
                    match tag {
                        Tag::PubKey(pubkey) => {
                            if let Some(relayer) = pubkey.relayer_url.as_ref() {
                                let relayer = relayer.to_ascii_lowercase();
                                if !relayer.is_empty() {
                                    if relayers.get_mut(&relayer).map(|x| *x += 1).is_none() {
                                        relayers.insert(relayer, 1);
                                    }
                                }
                            }
                        }
                        Tag::Event(tag) => {
                            if let Some(relayer) = tag.relayer_url.as_ref() {
                                let relayer = relayer.to_ascii_lowercase();
                                if !relayer.is_empty() {
                                    if relayers.get_mut(&relayer).map(|x| *x += 1).is_none() {
                                        relayers.insert(relayer, 1);
                                    }
                                }
                            }
                        }
                        _ => {}
                    }
                }
            }
            _ => {}
        }

        sqlx::query("UPDATE events SET processed = 2 WHERE id = ?")
            .bind(&value.id)
            .execute(&mut tx)
            .await?;
    }

    for (url, weight) in relayers.iter() {
        let _ = sqlx::query(
            r#"
            INSERT INTO relayers (url, weight) values(?, 0)
            "#,
        )
        .bind(url)
        .execute(&mut tx)
        .await;

        let _ = sqlx::query(
            r#"
            UPDATE relayers SET weight = weight + ? WHERE url = ?
            "#,
        )
        .bind(weight.to_string())
        .bind(url)
        .execute(&mut tx)
        .await;
    }

    tx.commit().await?;
    Ok(())
}

async fn process_events(conn: Pool<Sqlite>) -> Result<(), Error> {
    let mut tx = conn.begin().await?;
    let values = sqlx::query_as::<_, RawValue>(
        r#"SELECT id, event FROM events WHERE processed = 0 LIMIT 10000"#,
    )
    .fetch_all(&mut tx)
    .await?;

    for value in values {
        let event: Result<Event, _> = serde_json::from_str(&value.event);
        match event {
            Ok(event) => {
                for tag in event.inner.tags.iter() {
                    match tag {
                        Tag::PubKey(pubkey) => {
                            let _ = sqlx::query("INSERT INTO relationships (source_id, relates_to, type) values(?, ?, 2)")
                                .bind(&value.id)
                                .bind(pubkey.id.to_hex())
                                .execute(&mut tx)
                                .await;
                        }
                        Tag::Event(tag) => {
                            let _ = sqlx::query("INSERT INTO relationships (source_id, relates_to, type) values(?, ?, 1)")
                                .bind(&value.id)
                                .bind(tag.id.to_hex())
                                .execute(&mut tx)
                                .await;
                        }
                        _ => {}
                    }
                }
            }
            _ => {}
        }

        sqlx::query("UPDATE events SET processed = 1 WHERE id = ?")
            .bind(&value.id)
            .execute(&mut tx)
            .await?;
    }

    tx.commit().await?;
    Ok(())
}

async fn fetch_related_content(clients: Clients, conn: Pool<Sqlite>) -> Result<(), Error> {
    let mut tx = conn.begin().await?;
    sqlx::query(r#"DELETE FROM to_fetch"#)
        .execute(&mut tx)
        .await?;

    println!("Building to_fetch table...");

    sqlx::query(
        r#"
    INSERT INTO to_fetch SELECT * FROM (
        SELECT
            relates_to,
            count(*) total,
            1
        FROM
            relationships
        WHERE type = 1 AND relates_to NOT IN (SELECT id FROM events)
        GROUP BY relates_to
    ) s ORDER by total DESC
    "#,
    )
    .execute(&mut tx)
    .await?;

    sqlx::query(
        r#"
        INSERT INTO to_fetch SELECT * FROM (
            SELECT
                relates_to,
                count(*) total,
                2
            FROM
                relationships
            WHERE
                type = 2
                AND relates_to NOT IN (SELECT public_key FROM events WHERE kind = 0)
                AND relates_to NOT IN (SELECT id FROM to_fetch)
            GROUP BY relates_to
        ) s ORDER by total DESC
        "#,
    )
    .execute(&mut tx)
    .await?;

    tx.commit().await?;

    loop {
        let data_to_fetch = if let Ok(q) = sqlx::query_as::<_, ToFetch>(
            r#"
        SELECT
            id,
            type as typ
        FROM
            to_fetch
        ORDER BY refs DESC
        LIMIT 40
        "#,
        )
        .fetch_all(&conn)
        .await
        {
            q
        } else {
            println!("Database locked, retrying...");
            sleep(Duration::from_secs(1)).await;
            continue;
        };

        let mut public_keys = vec![];
        let mut ids = vec![];

        for to_fetch in data_to_fetch.into_iter() {
            if to_fetch.id.len() < 64 {
                continue;
            }

            let id: Addr = if let Ok(x) = to_fetch.id.try_into() {
                x
            } else {
                continue;
            };

            match to_fetch.typ {
                1 => {
                    ids.push(id);
                }
                2 => {
                    public_keys.push(id);
                }
                _ => {}
            }
        }

        let mut filters = vec![];

        if !ids.is_empty() {
            filters.push(Filter {
                ids: ids,
                ..Default::default()
            });
        }

        if !public_keys.is_empty() {
            filters.push(Filter {
                authors: public_keys,
                kinds: vec![
                    Kind::Metadata,
                    Kind::ShortTextNote,
                    Kind::Contacts,
                    Kind::Repost,
                    Kind::Reaction,
                    Kind::ZapRequest,
                    Kind::Zap,
                ],
                ..Default::default()
            });
        }

        if filters.is_empty() {
            break;
        }

        let subscription_id = SubscriptionId::default();
        println!("{} Fetching from related content", &*subscription_id,);

        clients
            .send(
                Subscribe {
                    subscription_id: subscription_id.clone(),
                    filters,
                    ..Default::default()
                }
                .into(),
            )
            .await;

        sleep(Duration::from_millis(15_000)).await;

        println!("{} Unsubscribing from related content", &*subscription_id);

        let _ = clients.send(Close(subscription_id).into()).await;

        let _ = sqlx::query("DELETE FROM to_fetch WHERE id IN (SELECT id FROM to_fetch ORDER BY refs DESC LIMIT 40)")
            .execute(&conn)
            .await;
    }

    Ok(())
}

#[tokio::main]
async fn main() {
    env_logger::init();
    let conn = SqlitePool::connect("sqlite://./db.sqlite").await.unwrap();
    let clients = Clients::default();

    let _ = query(
        r#"
    CREATE TABLE events(
        id varchar(64) not null primary key,
        public_key varchar(64) not null,
        kind int,
        event text,
        discovered_at datetime,
        discovered_by int,
        created_at datetime,
        processed INT DEFAULT 0
    );
    CREATE INDEX events_processed_index ON events (processed);
    CREATE INDEX events_public_key_index ON events (public_key, kind);
    CREATE INDEX events_kind_index ON events (kind);
    CREATE TABLE relationships (
        id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        source_id VARCHAR(64) NOT NULL,
        relates_to VARCHAR(64) NOT NULL,
        type INT
    );
    CREATE INDEX relationships_source_id_index ON relationships (source_id);
    CREATE TABLE relayers (
        id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        url VARCHAR(64) NOT NULL,
        weight INT NOT NULL DEFAULT '0'
    );
    CREATE UNIQUE INDEX url ON relayers (url);
    CREATE INDEX relayers_weight_index ON relayers (weight);
    CREATE TABLE to_fetch(id varchar(64) not null primary key, refs int not null default 0, type int not null default 1);
    CREATE INDEX refs ON to_fetch (refs);
    "#,
    )
    .execute(&conn)
    .await;

    let clients_for_worker = clients.clone();
    let conn_for_worker = conn.clone();
    tokio::spawn(async move {
        sleep(Duration::from_millis(35_000)).await;
        loop {
            let r1 =
                fetch_related_content(clients_for_worker.clone(), conn_for_worker.clone()).await;
            println!("Fetch related content {:?}", r1);
            sleep(Duration::from_millis(5_000)).await
        }
    });

    let conn_for_worker = conn.clone();
    tokio::spawn(async move {
        sleep(Duration::from_millis(55_000)).await;
        loop {
            let r1 = process_events(conn_for_worker.clone()).await;
            let r2 = discover_relayers(conn_for_worker.clone()).await;
            println!("Processed events {:?} {:?}", r1, r2);
            sleep(Duration::from_millis(5_000)).await;
        }
    });

    let clients_for_worker = clients.clone();
    tokio::spawn(async move {
        loop {
            sleep(Duration::from_millis(5_000)).await;
            println!(
                "Active connections: {}",
                clients_for_worker.check_active_connections()
            );
        }
    });

    let queue = Arc::new(RwLock::new(HashMap::<String, (Event, i64)>::new()));
    let conn_for_queue = conn.clone();
    let queue_for_worker = queue.clone();

    tokio::spawn(async move {
        loop {
            let to_persist = { queue_for_worker.read().clone() };

            if let Ok(mut tx) = conn_for_queue.begin().await {
                println!("Persisting {}", to_persist.len());
                let mut skip: u64 = 0;
                let mut processed: u64 = 0;
                for (_, (event, relayer_id)) in to_persist.iter() {
                    if let Ok(Some(_)) = query(r#"SELECT id FROM events WHERE id = ?"#)
                        .bind(event.id.to_string())
                        .fetch_optional(&mut tx)
                        .await
                    {
                        skip += 1;
                        continue;
                    }
                    let kind: u32 = event.inner.kind.into();
                    let created_at = event.inner.created_at.to_rfc3339();
                    let _ = query(
                        r#"
                        INSERT INTO events(id, public_key, kind, event, discovered_by, discovered_at, created_at)
                        VALUES(?, ?, ?, ?, ?, strftime('%Y-%m-%d %H:%M:%S','now'), ?)
                        "#,
                    )
                    .bind(event.id.to_string())
                    .bind(event.inner.public_key.to_string())
                    .bind(kind.to_string())
                    .bind(serde_json::to_string(&event).unwrap())
                    .bind(relayer_id)
                    .bind(created_at)
                    .execute(&mut tx)
                    .await;
                    processed += 1;
                }
                if let Ok(_) = tx.commit().await {
                    let mut queue = queue_for_worker.write();
                    for (id, _) in to_persist.iter() {
                        queue.remove(id);
                    }
                    println!(
                        "Persisted {}, skip = {} and (queue new size is {})",
                        processed,
                        skip,
                        queue.len()
                    );
                }
            }

            sleep(Duration::from_millis(60_000)).await;
        }
    });

    let relayers = sqlx::query_as::<_, Relayer>(
        r#"select id, url from relayers where url like 'wss://%.%/' order by weight desc limit 30"#,
    )
    .fetch_all(&conn)
    .await
    .expect("query")
    .iter()
    .map(|r| (r.url.clone(), r.id))
    .collect::<HashMap<String, i64>>();

    for (relayer, _) in relayers.iter() {
        let _ = clients.connect_to(&relayer).await;
    }

    loop {
        if let Some((msg, relayed_by)) = clients.recv().await {
            if let Response::Event(x) = msg {
                let event = x.event;
                let id = event.id.to_string();
                let mut q = queue.write();

                if !q.contains_key(&id) {
                    q.insert(
                        id,
                        (
                            event,
                            relayers.get(&relayed_by).cloned().unwrap_or_default(),
                        ),
                    );
                }
            }
        }
    }
}
